unit CDInfo;

interface

uses
  MMSystem, Windows;

type
  // declares my CD TOC
  TOC = record
    min: LongWord;
    sec: LongWord;
    frames: LongWord;
  end;

  TCDInfo = class(TObject)
  private
    mciOpenParms: MCI_OPEN_PARMS;
    mciStatusParms: MCI_STATUS_PARMS;
    procedure SendCommand(uMessage: UINT; dwParam1, dwParam2: LongWord);
    function GetStatusItem (AStatusItem: LongWord): LongWord;
  public
    constructor Create;
    destructor Destroy; override;
    procedure OpenDevice(Device: Char);
    function TotalTrackCount: LongWord;
    function TrackPosition(ATrack : integer): TOC;
    function TrackLength(ATrack : integer): TOC;
    function TotalCDLength: LongWord;
    function CDDBDiscId: LongWord;
    function CDDBDiscIDHex: string;
    class function TOCToFrames(ATOC: TOC): LongWord;
    class function MSFToTOC(AFrames: LongWord): TOC;
    class function FramesToTOC(AFrames: LongWord): TOC;
    class function TOCToString(ATOC : TOC): string;
  end;

implementation

uses
  SysUtils, Controls, Forms, Dialogs;

const
  CSearchDelta = 150;
{ TCDInfo }

constructor TCDInfo.Create;
begin
  inherited;
end;

destructor TCDInfo.Destroy;
begin
  // close MCI device
  SendCommand(MCI_CLOSE, 0, 0);
  inherited;
end;

procedure TCDInfo.OpenDevice(Device: Char);
var
  LmciSetParms : MCI_SET_PARMS;
  OldCursor: TCursor;
  Flags: DWord;
  szDriveName: Array[1..4] of Char;
begin
  OldCursor := Screen.Cursor;
  Screen.Cursor :=crHourglass;
  ZeroMemory(@mciOpenParms, sizeof(mciOpenParms));
  try
    Flags := MCI_OPEN_TYPE or MCI_OPEN_TYPE_ID or MCI_OPEN_SHAREABLE or MCI_OPEN_ELEMENT;
    // open MCI device
    mciOpenParms.lpstrDeviceType := PChar(MCI_DEVTYPE_CD_AUDIO);
    mciOpenParms.lpstrElementName := PChar(LowerCase(Device) + ':\0@');
    mciOpenParms.wDeviceID := 0;
    SendCommand(MCI_OPEN, Flags, LongWord(@mciOpenParms));
    // set time format to minute/second/frame (MSF) format
    LmciSetParms.dwTimeFormat := MCI_FORMAT_MSF;
    SendCommand(MCI_SET, MCI_SET_TIME_FORMAT, LongWord(@LmciSetParms));
  finally
    Screen.Cursor := OldCursor;
  end;
end;

procedure TCDInfo.SendCommand(uMessage: UINT; dwParam1, dwParam2: LongWord);
const
  CStrLength = 250;
var
  LError : MCIERROR;
  LErrorStr : string;
  OldCursor: TCursor;
begin
  OldCursor := Screen.Cursor;
  Screen.Cursor :=crHourglass;
  try
    LError := mciSendCommand(mciOpenParms.wDeviceID, uMessage, dwParam1, dwParam2);
    if LError <> 0 then begin
      SetLength(LErrorStr, CStrLength);
      mciGetErrorString(LError, PChar(LErrorStr), CStrLength);
      raise Exception.Create(LErrorStr);
    end;
  finally
    Screen.Cursor := OldCursor;
  end;
end;

function TCDInfo.TotalTrackCount : LongWord;
begin
  Result := GetStatusItem(MCI_STATUS_NUMBER_OF_TRACKS);
end;

function TCDInfo.TrackPosition(ATrack : integer) : TOC;
begin
  mciStatusParms.dwItem := MCI_STATUS_POSITION;
  mciStatusParms.dwTrack := ATrack;
  SendCommand(MCI_STATUS, MCI_STATUS_ITEM or MCI_TRACK, LongWord(@mciStatusParms));
  // save track position in MSF format
  Result := MSFToTOC(mciStatusParms.dwReturn);
end;

function TCDInfo.TrackLength(ATrack : integer) : TOC;
begin
  mciStatusParms.dwItem := MCI_STATUS_LENGTH;
  mciStatusParms.dwTrack := ATrack;
  SendCommand(MCI_STATUS, MCI_STATUS_ITEM or MCI_TRACK, LongWord(@mciStatusParms));
  // save track position in MSF format
  Result := MSFToTOC(mciStatusParms.dwReturn);
end;

class function TCDInfo.TOCToFrames(ATOC: TOC): LongWord;
begin
  Result := (ATOC.min * 60 * 75) + (ATOC.sec * 75) + ATOC.frames;
end;

class function TCDInfo.MSFToTOC(AFrames: LongWord): TOC;
begin
  Result.frames := MCI_MSF_FRAME(AFrames);
  Result.sec := MCI_MSF_SECOND(AFrames);
  Result.min := MCI_MSF_MINUTE(AFrames);
end;

function TCDInfo.TotalCDLength: LongWord;
var
  LLastTrack, LeadOutTrack : TOC;
  dwPos : LongWord;
begin
  // Compute lead-out track position //
  // get last track length
  LLastTrack := TrackLength(TotalTrackCount);
  // +1 to fix Windows MCI bug, according to cddb howto
  LLastTrack.frames := LLastTrack.frames + 1;
  // compute lead-out track position (in frame format)
  dwPos := TOCToFrames(TrackPosition(TotalTrackCount)) + TOCToFrames(LLastTrack);
  // save lead-out track position (in MSF format)
  LeadOutTrack := FramesToTOC(dwPos);
  // compute total CD length in seconds
  Result := ((LeadOutTrack.min * 60) + LeadOutTrack.sec) -
    ((TrackPosition(1).min * 60) + TrackPosition(1).sec);
end;

function TCDInfo.CDDBDiscIDHex: string;
begin
  Result := Format('%8.8x', [CDDBDiscId]);
end;

function TCDInfo.CDDBDiscId : LongWord;
var
  // general variables
  i: LongWord;
  // CD TOC information variables
  cdtoc: array[0..99] of TOC;
  // cddb discid computation variables
  j, n, cddb_sum : LongWord;
begin
  /////////////////
  // Read CD TOC //
  // for each track, get position (starting location)
  for i:= 1 to TotalTrackCount do begin
    cdtoc[i-1] := TrackPosition(i);
  end;
  /////////////////////
  // Compute DISC-ID //
  // For backward compatibility this algorithm must not change
  // cddb_discid
  i := 0; n := 0;
  while (i < TotalTrackCount) do begin
    // cddb_sum
    cddb_sum := 0;
    j := (cdtoc[i].min * 60) + cdtoc[i].sec;
    while (j > 0) do begin
      cddb_sum := cddb_sum + (j mod 10);
      j := j div 10;
    end;
    n := n + cddb_sum;
    Inc(i);
  end;
  // compute disc-id
  Result := ((n mod $ff) shl 24 or TotalCDLength shl 8 or TotalTrackCount);
end;

class function TCDInfo.FramesToTOC(AFrames : LongWord) : TOC;
begin
  Result.frames := AFrames mod 75;
  AFrames := AFrames div 75;
  Result.sec := AFrames mod 60;
  AFrames := AFrames div 60;
  Result.min := AFrames;
end;

class function TCDInfo.TOCToString(ATOC: TOC): string;
begin
  Result := IntToStr(ATOC.min) + ':' + Format('%.2d', [ATOC.sec]);
end;

function TCDInfo.GetStatusItem (AStatusItem: LongWord): LongWord;
begin
  mciStatusParms.dwItem := AStatusItem;
  SendCommand(MCI_STATUS, MCI_STATUS_ITEM, LongWord(@mciStatusParms));
  Result := mciStatusParms.dwReturn;
end;

end.
